"""Tools for communicating with HTChirp."""

import asyncio
import enum
import logging
import sys
import time
import traceback
from functools import wraps
from typing import Any, Callable, Coroutine, TypeVar

import htchirp  # type: ignore[import-untyped]
from htcondor import classad  # type: ignore[import-untyped]
from typing_extensions import ParamSpec

from .config import ENV

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")


class PilotStatus(enum.Enum):
    """A simple listing of statuses, useful for reporting & aggregation."""

    Started = enum.auto()
    RunningInitCommand = enum.auto()
    AwaitingFirstMessage = enum.auto()
    Tasking = enum.auto()
    PendingRemainingTasks = enum.auto()

    FatalError = enum.auto()
    Done = enum.auto()


HTCHIRP_ATTR_PREFIX = "HTChirpEWMSPilot"


class HTChirpEWMSPilotAttr(enum.Enum):
    """Organized list of attributes for chirping."""

    LastUpdatedTimestamp = enum.auto()  # always overwritten

    StartedTimestamp = enum.auto()
    Status = enum.auto()

    TasksTotal = enum.auto()
    PilotTasksFailed = enum.auto()
    TasksSuccess = enum.auto()

    PilotError = enum.auto()
    ErrorTraceback = enum.auto()


def _chirp(ctx: htchirp.HTChirp, _attr: HTChirpEWMSPilotAttr, _val: Any) -> None:
    """Chirp to either a job attr (default) or the job event log."""
    LOGGER.info(f"HTChirp -> {_attr.name} = {_val}")

    if ENV.EWMS_PILOT_HTCHIRP_DEST == "JOB_EVENT_LOG":
        # classad.quote escape new-lines, etc.
        ctx.ulog(f"{HTCHIRP_ATTR_PREFIX}{_attr.name}: {classad.quote(str(_val))}")
    elif ENV.EWMS_PILOT_HTCHIRP_DEST == "JOB_ATTR":
        # condor has built-in types (see below for strs)
        if isinstance(_val, (int, float)):
            # https://htcondor.readthedocs.io/en/latest/classads/classad-mechanism.html#composing-literals
            ctx.set_job_attr(f"{HTCHIRP_ATTR_PREFIX}{_attr.name}", str(_val))
        else:
            ctx.set_job_attr(
                f"{HTCHIRP_ATTR_PREFIX}{_attr.name}", classad.quote(str(_val))
            )
    else:
        raise ValueError(
            f"Invalid EWMS_PILOT_HTCHIRP_DEST: {ENV.EWMS_PILOT_HTCHIRP_DEST}"
        )


class Chirper:
    """Handle htchirp connection(s) and sending."""

    def __init__(self) -> None:
        self._conn = None
        self._backlog: dict[HTChirpEWMSPilotAttr, Any] = {}
        self._last_backlog_time = 0.0

    def _get_conn(self) -> htchirp.HTChirp:
        """Get chirp object, (re)establishing the connection if needed."""
        if self._conn:
            return self._conn

        try:  # checks if ".chirp.config" is present / provided a host and port
            self._conn = htchirp.HTChirp()
            self._conn.__enter__()  # type: ignore[attr-defined]
            return self._conn
        except Exception as e:
            LOGGER.error(f"HTChirp not available ({type(e).__name__}: {e})")
            raise

    def _reset_conn(self) -> None:
        self.close()

    def close(self, *args: Any) -> None:
        """Close the connection with the Chirp server."""
        if not self._conn:
            return
        try:
            self._conn.__exit__(*args)
        except Exception as e:
            LOGGER.error("chirping exit failed")
            LOGGER.exception(e)
        finally:
            self._conn = None

    async def chirp_backlog_until_done(self, total_time: int, sleep: int) -> float:
        """Call `chirp_backlog()` until backlog is all sent successfully.

        Total time may take a little bit longer than `total_time`,
        depending on the `sleep` value and the runtime of `chirp_backlog()`.

        Return the amount of time leftover from `total_time`.
        """
        if not ENV.EWMS_PILOT_HTCHIRP:
            return total_time

        start = time.time()

        while True:
            self.chirp_backlog()
            if (not self._backlog) or (time.time() - start >= total_time):
                break
            await asyncio.sleep(sleep)

        return max(0.0, total_time - (time.time() - start))  # remainder

    def chirp_backlog(self, is_rate_limited: bool = False) -> None:
        """Set all job attrs plus an additional attr -- a timestamp."""
        if not ENV.EWMS_PILOT_HTCHIRP:
            return

        if is_rate_limited and (
            time.time() - self._last_backlog_time
            < ENV.EWMS_PILOT_HTCHIRP_RATELIMIT_INTERVAL
        ):
            return

        # set HTChirpEWMSPilotLastUpdatedTimestamp & verify backlog
        self._backlog.pop(HTChirpEWMSPilotAttr.LastUpdatedTimestamp, None)
        if not self._backlog:
            return  # nothing to chirp
        now = int(time.time())
        self._backlog[HTChirpEWMSPilotAttr.LastUpdatedTimestamp] = now

        # chirp it all
        try:
            conn = self._get_conn()
            for bl_attr, bl_value in list(self._backlog.items()):
                _chirp(conn, bl_attr, bl_value)
                self._backlog.pop(bl_attr)  # wait to remove until success
        except Exception as e:
            LOGGER.error("chirping failed")
            LOGGER.exception(e)
            self._reset_conn()
        else:
            self._last_backlog_time = time.time()  # wait to set until all success

    def chirp_status(self, status: PilotStatus) -> None:
        """Invoke HTChirp, AKA send a status message to Condor."""
        if status == PilotStatus.Started:
            self._backlog[HTChirpEWMSPilotAttr.StartedTimestamp] = int(time.time())

        self._backlog[HTChirpEWMSPilotAttr.Status] = status.name
        self.chirp_backlog()

    def chirp_new_total(self, total: int) -> None:
        """Send a Condor Chirp signalling a new total of tasks handled.

        This chirp is enqueued (rate limited) and sent every X seconds.
        """
        if not total:
            # total can only increase -> can be inferred total=0 if attr is absent
            return

        self._backlog[HTChirpEWMSPilotAttr.TasksTotal] = total
        self.chirp_backlog(is_rate_limited=True)

    def chirp_new_success_total(self, total: int) -> None:
        """Send a Condor Chirp signalling a new total of succeeded task(s).

        This chirp is enqueued (rate limited) and sent every X seconds.
        """
        if not total:
            # total can only increase -> can be inferred total=0 if attr is absent
            return

        self._backlog[HTChirpEWMSPilotAttr.TasksSuccess] = total
        self.chirp_backlog(is_rate_limited=True)

    def chirp_new_failed_total(self, total: int) -> None:
        """Send a Condor Chirp signalling a new total of failed task(s).

        This chirp is enqueued (rate limited) and sent every X seconds.
        """
        if not total:
            # total can only increase -> can be inferred total=0 if attr is absent
            return

        self._backlog[HTChirpEWMSPilotAttr.PilotTasksFailed] = total
        self.chirp_backlog(is_rate_limited=True)

    def initial_chirp(self) -> None:
        """Send a Condor Chirp signalling that processing has started."""
        self.chirp_status(PilotStatus.Started)

    def error_chirp(self, exception: Exception) -> None:
        """Send a Condor Chirp signalling that processing ran into an error."""
        str_error = f"{type(exception).__name__}: {exception}"
        self._backlog[HTChirpEWMSPilotAttr.PilotError] = str_error

        if sys.version_info >= (3, 10):
            str_traceback = "".join(traceback.format_exception(exception))
        else:  # backwards compatibility
            # grabbed this from `logging.Logger._log()`
            if isinstance(exception, BaseException):
                exc_info = (type(exception), exception, exception.__traceback__)
            else:
                exc_info = sys.exc_info()
            str_traceback = "".join(traceback.format_exception(*exc_info))
        self._backlog[HTChirpEWMSPilotAttr.ErrorTraceback] = str_traceback

        self.chirp_backlog()


def async_htchirp_error_wrapper(
    func: Callable[P, Coroutine[Any, Any, T]]
) -> Callable[P, Coroutine[Any, Any, T]]:
    """Send Condor Chirp of any raised non-excepted exception."""

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            ret = await func(*args, **kwargs)
            return ret
        except Exception as e:
            chirper = Chirper()
            chirper.error_chirp(e)
            chirper.close()
            raise

    return wrapper



"""Logic for running a subprocess."""

import asyncio
import dataclasses as dc
import logging
import shutil
import subprocess
import sys
from pathlib import Path
from typing import TextIO

from ..config import ENV, PILOT_DATA_DIR, PILOT_DATA_HUB_DIR

LOGGER = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------


def get_last_line(fpath: Path) -> str:
    """Get the last line of the file."""
    with fpath.open() as f:
        line = ""
        for line in f:
            pass
        return line.rstrip()  # remove trailing '\n'


class PilotSubprocessError(Exception):
    """Raised when the subprocess terminates in an error."""

    def __init__(self, return_code: int, stderrfile: Path):
        super().__init__(
            f"Subprocess completed with exit code {return_code}: "
            f"{get_last_line(stderrfile)}"
        )


# --------------------------------------------------------------------------------------


class DirectoryCatalog:
    """Handles the naming and mapping logic for a task's directories."""

    @dc.dataclass
    class _ContainerBindMountDirPair:
        on_host: Path
        in_container: Path

    def __init__(self, name: str):
        """Directories are not pre-created; you must `mkdir -p` to use."""
        self._namebased_dir = PILOT_DATA_DIR / name

        # for inter-task/init storage: startup data, init container's output, etc.
        self.pilot_data_hub = self._ContainerBindMountDirPair(
            PILOT_DATA_HUB_DIR,
            PILOT_DATA_HUB_DIR,
        )

        # for persisting stderr and stdout
        self.outputs_on_host = self._namebased_dir / "outputs"

        # for message-based task i/o
        self.task_io = self._ContainerBindMountDirPair(
            self._namebased_dir / "task-io",
            PILOT_DATA_DIR / "task-io",
        )

    def assemble_bind_mounts(
        self,
        external_directories: bool = False,
        task_io: bool = False,
    ) -> str:
        """Get the docker bind mount string containing the wanted directories."""
        string = f"--mount type=bind,source={self.pilot_data_hub.on_host},target={self.pilot_data_hub.in_container} "

        if external_directories:
            string += "".join(
                f"--mount type=bind,source={dpath},target={dpath},readonly "
                for dpath in ENV.EWMS_PILOT_EXTERNAL_DIRECTORIES.split(",")
                if dpath  # skip any blanks
            )

        if task_io:
            string += f"--mount type=bind,source={self.task_io.on_host},target={self.task_io.in_container} "

        return string

    def rm_unique_dirs(self) -> None:
        """Remove all directories (on host) created for use only by this container."""
        shutil.rmtree(self._namebased_dir)  # rm -r


# --------------------------------------------------------------------------------------


def _dump_binary_file(fpath: Path, stream: TextIO) -> None:
    try:
        with open(fpath, "rb") as file:
            while True:
                chunk = file.read(4096)
                if not chunk:
                    break
                stream.buffer.write(chunk)
    except Exception as e:
        LOGGER.error(f"Error dumping subprocess output ({stream.name}): {e}")


class ContainerRunner:
    """A utility class to run a container."""

    def __init__(self, image: str, args: str, timeout: int | None):
        self.args = args
        self.timeout = timeout
        self.image = self._prepull_image(image)

    @staticmethod
    def _prepull_image(image: str) -> str:
        """Pull the image so it can be used in many tasks.

        Return the fully-qualified image name.
        """
        LOGGER.info(f"Pulling image: {image}")

        def _run(cmd: str):
            LOGGER.info(f"Running command: {cmd}")
            return subprocess.run(cmd, text=True, check=True, shell=True)

        match ENV._EWMS_PILOT_CONTAINER_PLATFORM.lower():

            case "docker":
                if ENV.CI:  # optimization during testing, images are *loaded* manually
                    return image
                _run(f"docker pull {image}")
                return image

            # NOTE: We are only are able to run unpacked directory format on condor.
            #       Otherwise, we get error: `code 255: FATAL:   container creation
            #       failed: image driver mount failure: image driver squashfuse_ll
            #       instance exited with error: squashfuse_ll exited: fuse: device
            #       not found, try 'modprobe fuse' first`
            #       See https://github.com/Observation-Management-Service/ewms-pilot/pull/86
            case "apptainer":
                if Path(image).is_dir():
                    LOGGER.info("OK: Apptainer image is already in directory format")
                    return image
                # assume non-specified image is docker -- https://apptainer.org/docs/user/latest/build_a_container.html#overview
                if "." not in image and "://" not in image:
                    # is not a blah.sif file (or other) and doesn't point to a registry
                    image = f"docker://{image}"
                # name it something that is recognizable -- and put it where there is enough space
                dir_image = (
                    f"{ENV._EWMS_PILOT_APPTAINER_BUILD_WORKDIR}/"
                    f"{image.replace('://', '_').replace('/', '_')}/"
                )
                # build (convert)
                _run(
                    # cd b/c want to *build* in a directory w/ enough space (intermediate files)
                    f"cd {ENV._EWMS_PILOT_APPTAINER_BUILD_WORKDIR} && "
                    f"apptainer build "
                    f"--fix-perms "
                    f"--sandbox {dir_image} "
                    f"{image}"
                )
                LOGGER.info(
                    f"Image has been converted to Apptainer directory format: {dir_image}"
                )
                return dir_image

            # ???
            case other:
                raise ValueError(
                    f"'_EWMS_PILOT_CONTAINER_PLATFORM' is not a supported value: {other}"
                )

    async def run_container(
        self,
        stdoutfile: Path,
        stderrfile: Path,
        mount_bindings: str,
        env_options: str,
        infile_arg_replacement: str = "",
        outfile_arg_replacement: str = "",
    ) -> None:
        """Run the container and dump outputs."""
        dump_output = ENV.EWMS_PILOT_DUMP_TASK_OUTPUT

        # insert in/out files *paths* into task_args
        inst_args = self.args
        if infile_arg_replacement:
            inst_args = inst_args.replace("{{INFILE}}", infile_arg_replacement)
        if outfile_arg_replacement:
            inst_args = inst_args.replace("{{OUTFILE}}", outfile_arg_replacement)

        # assemble command
        # NOTE: don't add to mount_bindings (WYSIWYG); also avoid intermediate structures
        match ENV._EWMS_PILOT_CONTAINER_PLATFORM.lower():
            case "docker":
                cmd = (
                    f"docker run --rm "
                    f"{mount_bindings} "
                    f"{env_options} "
                    f"{self.image} {inst_args}"
                )
            case "apptainer":
                cmd = (
                    f"apptainer run "
                    f"--containall --writable-tmpfs --no-eval "  # gets us close to docker functionality
                    f"{mount_bindings} "
                    f"{env_options} "
                    f"{self.image} {inst_args}"
                )
            case other:
                raise ValueError(
                    f"'_EWMS_PILOT_CONTAINER_PLATFORM' is not a supported value: {other}"
                )
        LOGGER.info(f"Running command: {cmd}")

        # run: call & check outputs
        try:
            with open(stdoutfile, "wb") as stdoutf, open(stderrfile, "wb") as stderrf:
                # await to start & prep coroutines
                proc = await asyncio.create_subprocess_shell(
                    cmd,
                    stdout=stdoutf,
                    stderr=stderrf,
                )
                # await to finish
                try:
                    await asyncio.wait_for(  # raises TimeoutError
                        proc.wait(),
                        timeout=self.timeout,
                    )
                except (TimeoutError, asyncio.exceptions.TimeoutError) as e:
                    # < 3.11 -> asyncio.exceptions.TimeoutError
                    raise TimeoutError(
                        f"subprocess timed out after {self.timeout}s"
                    ) from e

            LOGGER.info(f"Subprocess return code: {proc.returncode}")

            # exception handling (immediately re-handled by 'except' below)
            if proc.returncode:
                raise PilotSubprocessError(proc.returncode, stderrfile)

        except Exception as e:
            LOGGER.error(f"Subprocess failed: {e}")  # log the time
            dump_output = True
            raise
        finally:
            if dump_output:
                _dump_binary_file(stdoutfile, sys.stdout)
                _dump_binary_file(stderrfile, sys.stderr)
